#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pcap.h>
#include <signal.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <inttypes.h>

pcap_t             * handle = NULL;
struct sockaddr_in   dest;
int                  s;

void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet);
char * extract_tuple(const u_char * packet_ptr, bpf_u_int32 caplen);
void bailout(int signo);

int main(int argc, char *argv[]) {

    char *dev;
    struct bpf_program fp;
    char errbuf[PCAP_ERRBUF_SIZE];
    char filter_exp[] = "(dst port 22 or dst port 80 or dst portrange 40000-40050 or dst port 9090) and not host 127.0.0.1 and not host 10.22.2.2 and not host 10.22.2.3 and not host 10.22.2.4";
    bpf_u_int32 mask;
    bpf_u_int32 net;

    dev = "eth0";
    if (pcap_lookupnet(dev, &net, &mask, errbuf) == -1) {
        fprintf(stderr, "Couldn't get netmask for device %s: %s\n", dev, errbuf);
        return -1;
    }

    printf("Opening handle...\n");
    handle = pcap_open_live(dev, BUFSIZ, 1, 1000, errbuf);
    if (handle == NULL) {
        fprintf(stderr, "Couldn't open device %s: %s\n", dev, errbuf);
        return -1;
    }

    printf("Setting up filter...\n");
    if(pcap_compile(handle, &fp, filter_exp, 1, net) == -1) {
        fprintf(stderr, "Couldn't parse filter %s: %s\n", filter_exp, pcap_geterr(handle));
        return -1;
    }
    if(pcap_setfilter(handle, &fp) == -1) {
        fprintf(stderr, "Couldn't install filter %s: %s\n", filter_exp, pcap_geterr(handle));
        return -1;
    }

    printf("Setting up signals...\n");

    if(SIG_ERR == signal(SIGINT, bailout)) {
        fprintf(stderr, "Unable to set up signal for SIGINT\n");
        return -1;
    }
    if(SIG_ERR == signal(SIGTERM, bailout)) {
        fprintf(stderr, "Unable to set up signal for SIGTERM\n");
        return -1;
    }
    if(SIG_ERR == signal(SIGQUIT, bailout)) {
        fprintf(stderr, "Unable to set up signal for SIGQUIT\n");
        return -1;
    }

    printf("Setting up socket...\n");
    dest.sin_family      = AF_INET;
    dest.sin_port        = htons(50003);
    inet_pton(AF_INET, "10.22.2.4", &(dest.sin_addr));

    s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

    printf("Ready.\n");

    pcap_loop(handle, -1, got_packet, NULL);

    fprintf(stderr, "Program has exited pcap_loop abnormally\n");

    return -1;
}

void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet) {
    char * buf = extract_tuple(packet, header->caplen);
    if(NULL == buf) {
        fprintf(stderr, "Unable to extract tuple from packet! (likely neither UDP or TCP)\n");
        return;
    }

    if(sendto(s, buf, 64, 0, (struct sockaddr*) &dest, sizeof(dest)) < 0) {
        fprintf(stderr, "sendto failed on \"%s\"\n", buf);
    }
    free(buf);
}

// ip addrs are binary encoded still, not dot format
char * extract_tuple(const u_char * packet_ptr, bpf_u_int32 caplen) {
    char buf1[64], buf2[64], * src_ip, * dst_ip;
    struct in_addr temp;
    struct iphdr * i_hdr = NULL;
    char    * proto = NULL;
    uint16_t src_port, dst_port;
    
    memset(buf1, 0, 64);

    // advance past the ethernet header
    packet_ptr += sizeof(struct ether_header);
    caplen     -= sizeof(struct ether_header);

    // advance past the ip header
    i_hdr       = (struct iphdr*) packet_ptr;
    packet_ptr += 4*i_hdr->ihl;
    caplen     -= 4*i_hdr->ihl;

    temp.s_addr = i_hdr->saddr;
    src_ip = strdup(inet_ntoa(temp));
    
    temp.s_addr = i_hdr->daddr;
    dst_ip = strdup(inet_ntoa(temp));

    // TCP
    if(6 == i_hdr->protocol) {
        struct tcphdr * t_hdr = (struct tcphdr *) packet_ptr;
        src_port = ntohs(t_hdr->source);
        dst_port = ntohs(t_hdr->dest);
        proto = "tcp";
    }
    // UDP
    else if(17 == i_hdr->protocol) {
        struct udphdr * u_hdr = (struct udphdr *) packet_ptr;
        src_port = ntohs(u_hdr->source);
        dst_port = ntohs(u_hdr->dest);
        proto = "udp";
    }
    // Other
    else {
        free(src_ip);
        free(dst_ip);

        return NULL;
    }

    sprintf(buf1, "%s %s %s %" PRIu16 " %" PRIu16, proto, src_ip, dst_ip, src_port, dst_port);
    sprintf(buf2, "%-64s", buf1);

    free(src_ip);
    free(dst_ip);

    return strdup(buf2);
}

void bailout(int signo) {
    struct pcap_stat stats;

    if(pcap_stats(handle, &stats) >= 0) {
        printf("\n%d packets received\n", stats.ps_recv);
        printf("%d packets dropped\n\n", stats.ps_drop);
    }
    pcap_close(handle);

    exit(0);
}
