#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pcap.h>
#include <signal.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <inttypes.h>

#include <set>
#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>

using namespace std;

pcap_t             * handle = NULL;
struct sockaddr_in   dest;
int                  s;
set<string>          seen;

void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet);
void bailout(int signo);

int main(int argc, char *argv[]) {

    string dev;
    struct bpf_program fp;
    char errbuf[PCAP_ERRBUF_SIZE];
    char filter_exp[] = "(dst port 9090 or dst port 22 or dst portrange 40000-40050) and not host 127.0.0.1 and not host 10.22.2.2 and not host 10.22.2.3 and not host 10.22.2.4";
    bpf_u_int32 mask;
    bpf_u_int32 net;

    dev = "eth0";
    if (pcap_lookupnet(dev.c_str(), &net, &mask, errbuf) == -1) {
        cerr << "Couldn't get netmask for device " << dev << ": " << errbuf << "\n";
        return -1;
    }

    cout << "Opening handle...\n";
    handle = pcap_open_live(dev.c_str(), BUFSIZ, 1, 1000, errbuf);
    if (handle == NULL) {
        cerr << "Couldn't open device " << dev << ": " << errbuf << "\n";
        return -1;
    }

    cout << "Setting up filter...\n";
    if(pcap_compile(handle, &fp, filter_exp, 1, net) == -1) {
        cerr << "Couldn't parse filter " << filter_exp << ": " << pcap_geterr(handle) << "\n";
        return -1;
    }
    if(pcap_setfilter(handle, &fp) == -1) {
        cerr << "Couldn't install filter " << filter_exp << ": " << pcap_geterr(handle) << "\n";
        return -1;
    }

    cout << "Setting up signals...\n";

    if(SIG_ERR == signal(SIGINT, bailout)) {
        cerr << "Unable to set up signal for SIGINT\n";
        return -1;
    }
    if(SIG_ERR == signal(SIGTERM, bailout)) {
        cerr << "Unable to set up signal for SIGTERM\n";
        return -1;
    }
    if(SIG_ERR == signal(SIGQUIT, bailout)) {
        cerr << "Unable to set up signal for SIGQUIT\n";
        return -1;
    }

    cout << "Setting up socket...\n";
    dest.sin_family      = AF_INET;
    dest.sin_port        = htons(50003);
    inet_pton(AF_INET, "10.22.2.4", &(dest.sin_addr));

    s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

    cout << "Ready.\n";

    pcap_loop(handle, -1, got_packet, NULL);

    cerr << "Program has exited pcap_loop abnormally\n";

    return -1;
}

void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet) {
    char * src_ip, * dst_ip;
    struct in_addr temp;
    struct iphdr * i_hdr = NULL;
    string proto("");
    uint16_t src_port, dst_port;
    
    bpf_u_int32 caplen = header->caplen;

    // advance past the ethernet header
    packet += sizeof(struct ether_header);
    caplen -= sizeof(struct ether_header);

    // advance past the ip header
    i_hdr   = (struct iphdr*) packet;
    packet += 4*i_hdr->ihl;
    caplen -= 4*i_hdr->ihl;

    temp.s_addr = i_hdr->saddr;
    src_ip = strdup(inet_ntoa(temp));
    
    temp.s_addr = i_hdr->daddr;
    dst_ip = strdup(inet_ntoa(temp));

    // TCP
    if(6 == i_hdr->protocol) {
        struct tcphdr * t_hdr = (struct tcphdr *) packet;
        src_port = ntohs(t_hdr->source);
        dst_port = ntohs(t_hdr->dest);
        proto = "tcp";
    }
    // UDP
    else if(17 == i_hdr->protocol) {
        struct udphdr * u_hdr = (struct udphdr *) packet;
        src_port = ntohs(u_hdr->source);
        dst_port = ntohs(u_hdr->dest);
        proto = "udp";
    }
    // Other
    else {
        free(src_ip);
        free(dst_ip);

        return;
    }

    stringstream ss;
    char buf[64];
    ss << std::right << proto << " " << src_ip << " " << dst_ip << " " << src_port << " " << dst_port;
    sprintf(buf, "%64s", ss.str().c_str());
    string key(buf);
    if(seen.find(key) == seen.end()) {
        cout << ss.str() << "\n";
        if(sendto(s, key.c_str(), 64, 0, (struct sockaddr*) &dest, sizeof(dest)) < 0) {
            cerr << "sendto failed on \"" << key << "\"\n";
        }
        seen.insert(key);
    }

    free(src_ip);
    free(dst_ip);

    return;
}


void bailout(int signo) {
    struct pcap_stat stats;

    if(pcap_stats(handle, &stats) >= 0) {
        cout << "\n" << stats.ps_recv << "packets received\n";
        cout << stats.ps_drop << "packets dropped\n\n";
    }
    pcap_close(handle);

    exit(0);
}
