use strict;
use warnings;

use List::Util qw(max min);
use Math::NumberCruncher;
use POSIX qw(floor);

warn "Not handling router bandwidth\n";

my $cv = 1000; # seconds to milliseconds
my $dp = 9;    # number of decimal places

my $bcv = 8 / 1024; # byte to kbit

my %tcp_files;
my %udp_files;
my %router_files;

# aggregate files by test
foreach my $pcap_file (@ARGV) {
    $pcap_file =~ m/^\.\/([^\/]*)\//g;
    my $key = $1;
    if($pcap_file =~ m/router/i) {
        push(@{$router_files{$key}}, $pcap_file);
    }
    elsif($pcap_file =~ m/tcp/i) {
        push(@{$tcp_files{$key}}, $pcap_file);
    }
    elsif($pcap_file =~ m/udp/i) {
        push(@{$udp_files{$key}}, $pcap_file);
    }
    else {
        warn "\"$pcap_file\" an unknown type\n";
    }
}

trim_to_sender(\%tcp_files);
#pair_sender_receiver(\%udp_files);

tcp_analysis(\%tcp_files);
#udp_analysis(\%udp_files);


sub udp_analysis {
    my %pairs = %{shift @_};

    print "UDP ANALYSIS\n\n" if scalar keys %pairs;

    foreach my $test (keys %pairs) {
        my $lost_packets = 0;
        my %goodput      = ();
        my %throughput   = ();

        foreach my $id (keys %{$pairs{$test}}) {
            my @pair = @{$pairs{$test}->{$id}};

            my $cmd1 = 'tshark -r ' . $pair[0] . ' -T fields -e frame.time_epoch -e frame.len -e ip.len -e ip.hdr_len';
            my $cmd2 = 'tshark -r ' . $pair[1] . ' | wc -l';

            my @sent = split(/\n/, `$cmd1`);

            $lost_packets += scalar(@sent) - `$cmd2`;
            foreach my $line (@sent) {
                my @pieces = split(/\s+/, $line);
                $goodput{floor($pieces[0])}    += $bcv * ($pieces[2] - $pieces[3] - 8);
                $throughput{floor($pieces[0])} += $bcv * ($pieces[1]);
            }
        }

        my @good    = @goodput{sort { $a <=> $b } keys %goodput};
        my @through = @throughput{sort { $a <=> $b} keys %throughput};

        print "$test\n";
        print "Throughput Statistics\n";
        print "    Minimum         : " , min(@through)                                            , "\n";
        print "    Average         : " , Math::NumberCruncher::Mean(\@through),                   , "\n";
        print "    Median          : " , Math::NumberCruncher::Median(\@through, $dp)             , "\n";
        print "    Maximum         : " , max(@through)                                            , "\n";
        print "    Std. Dev        : " , Math::NumberCruncher::StandardDeviation(\@through, $dp)  , "\n";
        print "Goodput Statistics\n";
        print "    Minimum         : " , min(@good)                                               , "\n";
        print "    Average         : " , Math::NumberCruncher::Mean(\@good),                      , "\n";
        print "    Median          : " , Math::NumberCruncher::Median(\@good, $dp)                , "\n";
        print "    Maximum         : " , max(@good)                                               , "\n";
        print "    Std. Dev        : " , Math::NumberCruncher::StandardDeviation(\@good, $dp)     , "\n";
        print "Lost packets        : " , $lost_packets                                            , "\n";
        print                                                                                       "\n";
    }
}


sub tcp_analysis {
    my %files = %{shift @_};

    print "TCP ANALYSIS\n\n" if scalar keys %files;

    foreach my $test (sort keys %files) {
        my @rtt_vals        = ();
        my %goodput         = ();
        my %throughput      = ();
        my $retransmissions = 0;

        my $num_pcaps = 0;

        foreach my $pcap_file (@{$files{$test}}) {
            $num_pcaps++;
            $pcap_file =~ m/([^\/]*)_\d+\.pcap$/g;
            my ($src_ip) = split('_', $1);

            my $cmd1 = 'tshark -r ' . $pcap_file . ' -T fields -e tcp.analysis.ack_rtt | grep -v \'^$\'';
            my $cmd2 = 'tshark -r ' . $pcap_file . ' -R "tcp.analysis.retransmission or tcp.analysis.duplicate_ack or tcp.analysis.lost_segment or tcp.analysis.fast_retransmission" | wc -l';
            my $cmd3 = 'tshark -R "not tcp.analysis.ack_rtt and not (tcp.analysis.retransmission or tcp.analysis.duplicate_ack or tcp.analysis.lost_segment or tcp.analysis.fast_retransmission)" -r ' . $pcap_file . ' -T fields -e frame.time_epoch -e frame.len -e ip.len -e ip.hdr_len -e tcp.hdr_len';

            push(@rtt_vals, map { $cv * $_ } split(/\n/, `$cmd1`));
            $retransmissions += `$cmd2`;
   
            foreach my $line (split(/\n/, `$cmd3`)) {
                my @pieces = split(/\s+/, $line);
                $goodput{floor($pieces[0])}    += $bcv * ($pieces[2] - $pieces[3] - $pieces[4]);
                $throughput{floor($pieces[0])} += $bcv * ($pieces[1]);
            }
        }

        my @good    = @goodput{sort { $a <=> $b } keys %goodput};
        my @through = @throughput{sort { $a <=> $b} keys %throughput};

        print "$test\n";
        print "RTT Statistics\n";
        print "    Minimum         : " , min(@rtt_vals)                                           , "\n";
        print "    Average         : " , Math::NumberCruncher::Mean(\@rtt_vals),                  , "\n";
        print "    Median          : " , Math::NumberCruncher::Median(\@rtt_vals, $dp)            , "\n";
        print "    Maximum         : " , max(@rtt_vals)                                           , "\n";
        print "    Std. Dev        : " , Math::NumberCruncher::StandardDeviation(\@rtt_vals, $dp) , "\n";
        print "    Count           : " , scalar(@rtt_vals)                                        , "\n";
        print "Throughput Statistics\n";
        print "    Minimum         : " , min(@through)                                            , "\n";
        print "    Average         : " , Math::NumberCruncher::Mean(\@through) / $num_pcaps,      , "\n";
        print "    Median          : " , Math::NumberCruncher::Median(\@through, $dp)/$num_pcaps  , "\n";
        print "    Maximum         : " , max(@through)                                            , "\n";
        print "    Std. Dev        : " , Math::NumberCruncher::StandardDeviation(\@through, $dp)  , "\n";
        print "Goodput Statistics\n";
        print "    Minimum         : " , min(@good)                                               , "\n";
        print "    Average         : " , Math::NumberCruncher::Mean(\@good)/$num_pcaps            , "\n";
        print "    Median          : " , Math::NumberCruncher::Median(\@good, $dp)/$num_pcaps     , "\n";
        print "    Maximum         : " , max(@good)                                               , "\n";
        print "    Std. Dev        : " , Math::NumberCruncher::StandardDeviation(\@good, $dp)     , "\n";
        print "Retransmissions     : " , $retransmissions                                         , "\n";
        print "Total Packets Sent  : " , scalar(@rtt_vals)                                        , "\n";
        print "Number of pcaps     : " , $num_pcaps                                               , "\n";
        print                                                                                       "\n";
    }
}

sub trim_to_sender {
    my $files = shift;

    foreach my $key (keys %{$files}) {
        my @trimmed_pcap_files = ();
        foreach (@{$files->{$key}}) {
            m/([^\/]*)_\d+\.pcap$/g;
            my @ips = split('_', $1);
            next if $ips[0] gt $ips[1];
            push(@trimmed_pcap_files, $_);
        }
        $files->{$key} = \@trimmed_pcap_files;
    }
}

sub pair_sender_receiver {
    my $files = shift;
    my %pairs = ();

    foreach my $key (keys %{$files}) {
        foreach (@{$files->{$key}}) {
            m/([^\/]*)_\d+\.pcap$/g;
            my $id = $1;
            my @ips = split('_', $id);
            if($ips[0] lt $ips[1]) {
                unshift(@{$pairs{$key . '-' . $id}}, $_);
            }
            else {
                $id = $ips[1] . '_' . $ips[0] if $ips[0] gt $ips[1];
                push(@{$pairs{$key . '-' . $id}}, $_);
            }
        }
        $files->{$key} = \%pairs;
    }
}
